==> moving_average-0.2.0.tar.gz: lib/src/moving_average_base.dart <==
/*
 * Copyright (c) 2019. Peter Scully
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 *  Neither the name of the copyright holder nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import 'dart:math';

enum AverageType { simple, weighted, exponential }

// TODO: What about taking a List<T1> and returning a T2?
// e.g. List<WeighIn> => DataPoint
class MovingAverage<T> {
  final AverageType averageType;
  final int windowSize;
  final double factor;
  final bool partialStart;
  final num Function(T) getValue;
  final T Function(List<T>, num newValue) add;

  MovingAverage({
    this.averageType = AverageType.simple,
    this.windowSize = 1,
    this.factor = 1.0,
    this.partialStart = false,
    this.getValue,
    this.add,
  });

  // TODO: Second constructor for Exponential Moving Average
  // TODO: Consider 2 factories that instantiates separate classes for simple/weighted, and exponential?
  // TODO: Consider simplified constructors for List<num> => num

  List<T> call(List<T> data) {
    var result = <T>[];
    num Function(List<num>) averageFunc;

    if (averageType == AverageType.exponential) {
      result.addAll(_exponential(data, factor));
    } else {
      if (averageType == AverageType.simple) {
        averageFunc = _mean;
      } else if (averageType == AverageType.weighted) {
        averageFunc = _weightedMean;
      }

      final subLists = _window(data, windowSize, partialStart: partialStart);
      for (var l in subLists) {
        final v = l.map(getValue).toList();
        final m = averageFunc(v);
        final r = add(l, m);
        result.add(r);
      }
    }

    return result;
  }

  /// Calculate the exponential moving average of data, with a weight factor of
  /// factor.
  List<T> _exponential(List<T> data, double factor) {
    final result = <T>[];
    if (data == null || data.isEmpty) {
      return result;
    }

    var prev = getValue(data.first);
    for (var i = 0; i < data.length; i++) {
      prev = prev * (1.0 - factor) + getValue(data[i]) * factor;
      result.add(add([data[i]], prev));
    }

    return result;
  }

  /// Calculate the mean (average) of data.
  num _mean(List<num> data) =>
      data.fold(
        0,
        (num a, num b) => a + b,
      ) /
      data.length;

  /// Calculate the weighted mean of data.
  num _weightedMean(List<num> data) {
    num sum = 0;
    for (var i = 0; i < data.length; i++) {
      sum += data[i] * (i + 1);
    }
    return sum / (data.length * (data.length + 1) / 2);
  }

  /// Return a list of sublists of data of size size.
  /// If partialStart is true then sublists of size 1,2,...n at the start of data will also be included.
  /// If partialEnd is true then sublists of size 1,2,...n at the end of data will also be included.
  Iterable<List<T>> _window(List<T> data, int size,
      {bool partialStart = false, bool partialEnd = false}) sync* {
    var start = 0;
    var end = 0;
    if (partialStart) {
      end = 1;
      start = end - size;
    } else {
      start = 0;
      end = size;
    }
    while (end <= data.length) {
      yield data.sublist(max(start, 0), end);
      start++;
      end++;
    }
    return;
  }
}

==> moving_average-0.2.1.tar.gz: lib/src/moving_average_base.dart <==
/*
 * Copyright (c) 2019. Peter Scully
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 *  Neither the name of the copyright holder nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import 'dart:math';

enum AverageType { simple, weighted, exponential }

// TODO: What about taking a List<T1> and returning a T2?
// e.g. List<WeighIn> => DataPoint

/// A class to calculate simple, weighted or exponential moving averages of
/// lists of objects.
///
/// [averageType] is the type of calculation to use.
/// [windowSize] is the window size used when calculating simple and weighted
/// moving averages.
/// [partialStart] is used with simple or weighted moving averages to control
/// whether or not to include calculations on incomplete windows at the start of
/// the data.
/// [factor] is the value used when calculating exponential moving averages.  It
/// should be constrained by 0.0 <= factor <= 1.0.
/// [getValue] is a function to take a value T and return a number,
/// e.g. the y coordinate of a point.
/// [add] is a function to take a list of inputs, and an average value, and
/// create a new data point from the data points and the new value.  Examples
/// include using the middle x value of a set of points for the x value of the
/// resulting point, and the new average for the y value.
///
/// For calculations on lists of numbers, not objects, use the following for
/// getValue and add:
/// ```dart
/// getValue: (num n) => n,
/// add: (List<num> data, num value) => value,
/// ````
class MovingAverage<T> {
  final AverageType averageType;
  final int windowSize;
  final bool partialStart;
  final double factor;
  final num Function(T) getValue;
  final T Function(List<T> data, num newValue) add;

  /// Instantiate the class with the gven parameters
  MovingAverage({
    this.averageType = AverageType.simple,
    this.windowSize = 1,
    this.partialStart = false,
    this.factor = 1.0,
    this.getValue,
    this.add,
  });

  // TODO: Second constructor for Exponential Moving Average
  // TODO: Consider 2 factories that instantiates separate classes for
  // simple/weighted, and exponential?
  // TODO: Consider simplified constructors for List<num> => num

  /// Invoke the configured MovingAverage calculator on data.
  List<T> call(List<T> data) {
    var result = <T>[];
    num Function(List<num>) averageFunc;

    if (averageType == AverageType.exponential) {
      result.addAll(_exponential(data, factor));
    } else {
      if (averageType == AverageType.simple) {
        averageFunc = _mean;
      } else if (averageType == AverageType.weighted) {
        averageFunc = _weightedMean;
      }

      final subLists = _window(data, windowSize, partialStart: partialStart);
      for (var l in subLists) {
        final v = l.map(getValue).toList();
        final m = averageFunc(v);
        final r = add(l, m);
        result.add(r);
      }
    }

    return result;
  }

  /// Calculate the exponential moving average of data, with a weight factor of
  /// factor.
  List<T> _exponential(List<T> data, double factor) {
    final result = <T>[];
    if (data == null || data.isEmpty) {
      return result;
    }

    var prev = getValue(data.first);
    for (var i = 0; i < data.length; i++) {
      prev = prev * (1.0 - factor) + getValue(data[i]) * factor;
      result.add(add([data[i]], prev));
    }

    return result;
  }

  /// Calculate the mean (average) of data.
  num _mean(List<num> data) =>
      data.fold(
        0,
        (num a, num b) => a + b,
      ) /
      data.length;

  /// Calculate the weighted mean of data.
  num _weightedMean(List<num> data) {
    num sum = 0;
    for (var i = 0; i < data.length; i++) {
      sum += data[i] * (i + 1);
    }
    return sum / (data.length * (data.length + 1) / 2);
  }

  /// Return a list of sublists of data of size size.
  /// If partialStart is true then sublists of size 1,2,...n at the start of data will also be included.
  /// If partialEnd is true then sublists of size 1,2,...n at the end of data will also be included.
  Iterable<List<T>> _window(List<T> data, int size,
      {bool partialStart = false, bool partialEnd = false}) sync* {
    var start = 0;
    var end = 0;
    if (partialStart) {
      end = 1;
      start = end - size;
    } else {
      start = 0;
      end = size;
    }
    while (end <= data.length) {
      yield data.sublist(max(start, 0), end);
      start++;
      end++;
    }
    return;
  }
}
